'use strict'

var test = require('tape')
var terminus = require('terminus')
var Routing = require('../router')
var events = require('events')
var EE = events.EventEmitter
var bus = new EE()
var mock = require('./mocks/mock')

var mocksReady = 0
function mocksReadyCB () {
  mocksReady++
  if (mocksReady === 2) runTestsEtcd()
}

var mock1 = mock()
var mock2 = mock()
mock1.on('ready', mocksReadyCB)
mock2.on('ready', mocksReadyCB)

function runTestsEtcd () {
  const etcd1 = '127.0.0.1:' + mock1.etcdPort
  const etcd2 = '127.0.0.1:' + mock2.etcdPort

  runTests(etcd1)
  runTests(etcd1 + ',' + etcd2)

  test('teardown', function (t) {
    // This sucks because you can't `test.only`
    // but tape doesn't have a teardown facility
    console.log('mock1.requestsCount: ' + mock1.requestsCount)
    console.log('mock2.requestsCount: ' + mock2.requestsCount)
    mock1.stop()
    mock2.stop()
    t.ok(mock1.requestsCount > 0, 'expecting etcd mock 1 to have requests')
    t.ok(mock2.requestsCount > 0, 'expecting etcd mock 2 to have requests')
    t.end()
  })
}

function runTests (etcdHosts) {
  var options = {
    etcd_host: etcdHosts,
    listener: bus,
    denied: ['foo'],
    broadcast_approved: ['ping']
  }
  var router = new Routing(options)

  function collectify (testfn) {
    return function collectResults (err, stream) {
      if (stream == null) {
        return testfn(err)
      }
      let collector = terminus.concat(function collect (content) {
        // un newline-delim the JSON into an array of objects
        let results = content.toString().trim().split('\n').map(function (line) {
          return JSON.parse(line)
        })
        testfn(err, results)
      })
      stream.on('error', testfn)
      collector.on('error', testfn)
      stream.pipe(collector)
    }
  }

  test('list all', function (t) {
    function checkResults (err, results) {
      t.notOk(err, 'No error')
      t.equal(results.length, 3, 'Results have all of the expected agents')
      results.forEach(function (host) {
        t.ok(host.id != null, 'has an id')
        t.ok(host.address, 'has an address')
        t.ok(host.app, 'has an app')
      })
      t.end()
    }
    router.list(null, {}, collectify(checkResults))
  })

  test('list one app', function (t) {
    function checkResults (err, results) {
      t.notOk(err, 'No error')
      t.equal(results.length, 2, 'Results have all of the expected agents')
      results.forEach(function (host) {
        t.ok(host.id > 0, 'has an id')
        t.ok(host.address, 'has an address')
        t.equal(host.app, 'mock', 'has correct app')
      })
      t.end()
    }
    router.list(null, {app: 'mock'}, collectify(checkResults))
  })

  test('broadcast command all', function (t) {
    function checkResults (err, results) {
      t.notOk(err, 'No error')
      t.equal(results.length, 3, 'Results have all of the expected agents')
      results.forEach(function (host) {
        if (host.meta.id === '0') {
          t.equals(host.error, 'NOPE (error)', 'broken host replies as error')
          t.notOk(host.reply, 'No reply when there is an error')
          t.equals(host.meta.app, 'broke', 'app is correct')
        } else {
          t.notOk(host.error, 'no error')
          t.ok(host.meta.id > 0, 'id is present & expected')
          t.equals(host.reply, 'PONG ping', 'reply is correct')
          t.equals(host.meta.app, 'mock', 'app is correct')
          t.equals(host.meta.pid, process.pid, 'pid is correct')
          t.ok(host.meta.address, 'has an address')
          t.equals(host.meta.hostname, 'foobar.hostname', 'has a hostname')
        }
      })
      t.end()
    }
    router.proxy(null, {command: 'ping'}, collectify(checkResults))
  })

  test('broadcast command one app', function (t) {
    function checkResults (err, results) {
      t.notOk(err, 'No error')
      t.equal(results.length, 2, 'Results have all of the expected agents')
      results.forEach(function (host) {
        t.notOk(host.error, 'no error')
        t.ok(host.meta.id > 0, 'id is present & expected')
        t.equals(host.reply, 'PONG ping', 'reply is correct')
        t.equals(host.meta.app, 'mock', 'app is correct')
      })
      t.end()
    }
    router.proxy(null, {command: 'ping', app: 'mock'}, collectify(checkResults))
  })

  test('broadcast command with nobody home', function (t) {
    function checkResults (err, results) {
      t.notOk(err, 'No error')
      t.notOk(results, 'No results')
      t.end()
    }
    router.proxy(null, {command: 'ping', app: 'nope'}, collectify(checkResults))
  })

  test('status error in reply', function (t) {
    function checkResults (err, results) {
      t.ok(err instanceof Error, 'err is an Error')
      t.notOk(results, 'no results (error)')
      t.equals(err.message, '"NOPE (error)"')
      t.end()
    }
    router.proxy(null, {command: 'ping', app: 'broke', id: 0}, collectify(checkResults))
  })

  test('single agent proxy', function (t) {
    function checkResults (err, results) {
      t.notOk(err, 'No error')
      t.equal(results.length, 1, 'Results have all of the expected agents')
      t.equal(results[0], 'PONG ping', 'correct reply')
      t.end()
    }
    router.proxy(null, {command: 'ping', app: 'mock', id: 2}, collectify(checkResults))
  })

  test('denied command', function (t) {
    function checkResults (err, results) {
      t.ok(err instanceof Error, 'error is an error')
      t.equal(err.message, 'Refusing to send denied command foo', 'This command is denied')
      t.notOk(results, 'no results')
      t.end()
    }
    router.proxy(null, {command: 'foo', app: 'mock', id: 2}, collectify(checkResults))
  })

  test('id with no app', function (t) {
    function checkResults (err, results) {
      t.ok(err instanceof Error, 'error is an error')
      t.equal(err.message, '`app` and `id` are both required for lookup', 'id and app required')
      t.notOk(results, 'no results')
      t.end()
    }
    router.proxy(null, {command: 'ping', id: 2}, collectify(checkResults))
  })

  test('single agent proxy non-broadcastable command', function (t) {
    function checkResults (err, results) {
      t.notOk(err, 'No error')
      t.equal(results.length, 1, 'Results have all of the expected agents')
      t.equal(results[0], 'PONG zzz', 'correct reply')
      t.end()
    }
    router.proxy(null, {command: 'zzz', app: 'mock', id: 2}, collectify(checkResults))
  })

  test('single agent proxy non-broadcastable command without id', function (t) {
    function checkResults (err, results) {
      t.ok(err instanceof Error, 'error is an error')
      t.equal(err.message, 'Command \'zzz\' is not broadcast approved, id is required.', 'This command is cannot be broadcast')
      t.notOk(results, 'no results')
      t.end()
    }
    router.proxy(null, {command: 'zzz', app: 'mock'}, collectify(checkResults))
  })

  test('send data (non-broadcast)', function (t) {
    function checkResults (err, results) {
      t.notOk(err, 'No error')
      t.equal(results.length, 1, 'Results have all of the expected agents')
      t.equal(results[0], 'PONG zzz <abc>', 'correct reply')
      t.end()
    }
    router.proxy(null, {command: 'zzz', app: 'mock', id: 2, data: 'abc'}, collectify(checkResults))
  })

  test('send data (broadcast)', function (t) {
    function checkResults (err, results) {
      t.notOk(err, 'no error')
      t.equal(results.length, 2, 'Results have all of the expected agents')
      results.forEach(function (host) {
        t.notOk(host.error, 'no error')
        t.ok(host.meta.id > 0, 'id is present & expected')
        t.equals(host.reply, 'PONG ping <abc>', 'reply is correct')
        t.equals(host.meta.app, 'mock', 'app is correct')
      })
      t.end()
    }
    router.proxy(null, {command: 'ping', app: 'mock', data: 'abc'}, collectify(checkResults))
  })

  test('no command', function (t) {
    function checkResults (err, results) {
      t.ok(err instanceof TypeError, 'error is an error')
      t.equal(err.message, 'Missing required command!', 'command is required for proxy')
      t.notOk(results, 'no results')
      t.end()
    }
    router.proxy(null, {}, collectify(checkResults))
  })

  test('custom timeout (test takes 1 second)', function (t) {
    function checkResults (err, results) {
      t.ok(err instanceof Error, 'error is an error')
      t.equal(err.message, 'Client request timeout', 'slow request times out')
      t.notOk(results, 'no results')
      t.end()
    }
    router.proxy(null, {command: 'slow', app: 'mock', id: 1, timeout: 1000}, collectify(checkResults))
  })

  test('timeout (test takes 10 seconds)', function (t) {
    function checkResults (err, results) {
      t.ok(err instanceof Error, 'error is an error')
      t.equal(err.message, 'Client request timeout', 'slow request times out')
      t.notOk(results, 'no results')
      t.end()
    }
    router.proxy(null, {command: 'slow', app: 'mock', id: 1}, collectify(checkResults))
  })
}
